header "post_include_hpp" {
#include <basicnodes/basic_forwarders.h>
#include <suifnodes/suif_forwarders.h>
#include <esuifnodes/esuif_forwarders.h>
#include <common/lstring.h>
#include <common/i_integer.h>
#include <antlr/Token.hpp>
#include "LineAST.hpp"
#include <stdarg.h>
class SuifEnv;
}

header "post_include_cpp" {
}

header {

/**********************************************************************

The c2s frontend:

For 

void main() { int a; }

FileSetBlock.external_symbol_table:

  built-in C types: a BooleanType, an unsigned IntegerType, a signed
  IntegerType, etc.  None have names.

  CProcedureType for the "void main()" function, unnamed.
  A QualifiedType pointing to the signed integer NumericType;
  ProcedureSymbol for "main", named.

FileSetBlock.file_set_symbol_table:

  empty

FileBlocks contains a single FileBlock

The FileBLock contains a single ProcedureDefinition

ProcedureDefinition.symbolTable contains the "a" variable symbol whose 
type is the QualifiedType for integer in the external_symbol_table.

The explicit_super_scope is symbol_table for the FileBlock, which is
empty.

VariableDefintions appear in FileBlock.definition_block for
global variables, not automatic variables.

Initialization of automatic variables is transformed into
StoreVariableStatements at the beginning of the procedure.

***********************************************************************/

#include <basicnodes/basic.h>
#include <basicnodes/basic_factory.h>
#include <suifnodes/suif.h>
#include <suifnodes/suif_factory.h>
#include <esuifnodes/esuif.h>
#include <esuifnodes/esuif_factory.h>

LString EsterelTreeParser::inputAnnote = "input";
LString EsterelTreeParser::outputAnnote = "output";
LString EsterelTreeParser::inputoutputAnnote = "inputoutput";
LString EsterelTreeParser::returnAnnote = "return";
LString EsterelTreeParser::sensorAnnote = "sensor";
LString EsterelTreeParser::localAnnote = "local";
LString EsterelTreeParser::integerTypeName = "integer";
LString EsterelTreeParser::booleanTypeName = "boolean";
LString EsterelTreeParser::stringTypeName = "string";
LString EsterelTreeParser::floatTypeName = "float";
LString EsterelTreeParser::doubleTypeName = "double";
LString EsterelTreeParser::lineAnnote = "line";
LString EsterelTreeParser::constantQualifier = "const";
LString EsterelTreeParser::tickSignalName = "tick";
LString EsterelTreeParser::voidTypeName = "void";

BuiltinFunctionSymbol *EsterelTreeParser::get_builtin_function(Context *c,
							       LString name)
{
  suif_assert(c);
  suif_assert(c->module);
  SymbolTableObject *sto =
    c->module->get_builtin_symbols()->lookup_lookup_table(name);
  if (sto) {
    return to<BuiltinFunctionSymbol>(sto);
  }
  BuiltinFunctionSymbol *bfs = create_builtin_function_symbol(suif_env, 0);
  c->module->get_builtin_symbols()->add_symbol(name, bfs);
  return bfs;
}

QualifiedType* EsterelTreeParser::get_const_type(Context *c,
						 DataType *type)
{
  for ( Iter<SymbolTableObject *> iter =
	  c->module->get_symbol_table()->get_symbol_table_object_iterator() ;
	iter.is_valid() ; iter.next() ) {
    SymbolTableObject *sto = iter.current();
    if ( is_a<QualifiedType>(sto) ) {
      QualifiedType *qt = to<QualifiedType>(sto);
      if (qt->get_base_type() == type &&
	  qt->has_qualification_member(constantQualifier))
        return qt;
    }
  }
  QualifiedType *qt = create_qualified_type(suif_env, type);
  qt->append_qualification(constantQualifier);
  qt->set_name(type->get_name());
  c->variables->append_symbol_table_object(qt);
  return qt;
}

VariableSymbol *EsterelTreeParser::find_variable(Context *c,
						 RefLineAST symbol)
{
  LString name(symbol->getText().c_str());
  for (SymbolTable *st = c->variables ;
       st ; st = st->get_explicit_super_scope()) {
    SymbolTableObject * sto = st->lookup_lookup_table(name);
    if (sto) return to<VariableSymbol>(sto);
  }
  report_error(symbol, "undeclared variable `%s'", name.c_str());

  // Create a fictious integer variable as a placeholder

  VariableSymbol *vs = create_variable_symbol(suif_env, qIntegerType, name);
  return vs;
}

TrapSymbol *EsterelTreeParser::find_trap(Context *c,
					 RefLineAST symbol)
{
  LString name(symbol->getText().c_str());
  for (SymbolTable *st = c->traps ;
       st ; st = st->get_explicit_super_scope()) {
    SymbolTableObject * sto = st->lookup_lookup_table(name);
    if (sto) return to<TrapSymbol>(sto);
  }
  report_error(symbol, "undeclared trap `%s'", name.c_str());

  // Create a fictious pure exception as a placeholder

  TrapSymbol *ts = create_trap_symbol(suif_env, qVoidType, name);
  return ts;
}

SignalSymbol *EsterelTreeParser::find_signal(Context *c,
					     RefLineAST symbol)
{
  LString name(symbol->getText().c_str());
  for (SymbolTable *st = c->signals ;
       st ; st = st->get_explicit_super_scope()) {
    SymbolTableObject * sto = st->lookup_lookup_table(name);
    if (sto) return to<SignalSymbol>(sto);
  }
  report_error(symbol, "undeclared signal `%s'", name.c_str());

  // Create a fictious integer signal as a placeholder

  SignalSymbol *ss = create_signal_symbol(suif_env, qIntegerType, name);
  return ss;
}

Expression* EsterelTreeParser::build_boolean_binary(RefLineAST token,
						    Expression *e1,
						    Expression *e2)
{
  suif_assert(e1);
  suif_assert(e2);
  LString op(token->getText().c_str());
  Expression *e =
    create_binary_expression(suif_env, e1->get_result_type(), op, e1, e2);

  DataType *dt1 = e1->get_result_type();
  DataType *dt2 = e2->get_result_type();
  suif_assert(dt1);
  suif_assert(dt2);
  if ( dt1 != dt2 || !is_a<BooleanType>(dt1) ) {
    report_error(token, "arguments of `%s' must both be boolean", op.c_str());
    report_warning(token, "left argument type is `%s'", dt1->get_name().c_str());
    report_warning(token, "right argument type is `%s'", dt2->get_name().c_str());
  }
  return e;
}

Expression* EsterelTreeParser::build_numeric_binary(RefLineAST token,
						    Expression *e1,
						    Expression *e2)
{
  suif_assert(e1);
  suif_assert(e2);
  LString op(token->getText().c_str());
  Expression *e =
    create_binary_expression(suif_env, e1->get_result_type(), op, e1, e2);

  DataType *dt1 = e1->get_result_type();
  DataType *dt2 = e2->get_result_type();
  suif_assert(dt1);
  suif_assert(dt2);
  if ( dt1 != dt2 || !(is_a<IntegerType>(dt1) ||
		       is_a<FloatingPointType>(dt1)) ) {
    report_error(token, "arguments of `%s' must be numeric and match", op.c_str());
    report_warning(token, "left argument type is `%s'", dt1->get_name().c_str());
    report_warning(token, "right argument type is `%s'", dt2->get_name().c_str());
  }
  return e;
}

Expression* EsterelTreeParser::build_equality_binary(RefLineAST token,
						     Expression *e1,
						     Expression *e2)
{
  suif_assert(e1);
  suif_assert(e2);
  LString op(token->getText().c_str());
  Expression *e =
    create_binary_expression(suif_env, booleanType, op, e1, e2);

  DataType *dt1 = e1->get_result_type();
  DataType *dt2 = e2->get_result_type();
  suif_assert(dt1);
  suif_assert(dt2);
  if ( dt1 != dt2 ) {
    report_error(token, "type of arguments of `%s' must match", op.c_str());
    report_warning(token, "left argument type is `%s'", dt1->get_name().c_str());
    report_warning(token, "right argument type is `%s'", dt2->get_name().c_str());
  }
  return e;
}

Expression* EsterelTreeParser::build_comparison_binary(RefLineAST token,
						       Expression *e1,
						       Expression *e2)
{
  suif_assert(e1);
  suif_assert(e2);
  LString op(token->getText().c_str());
  Expression *e =
    create_binary_expression(suif_env, booleanType, op, e1, e2);

  DataType *dt1 = e1->get_result_type();
  DataType *dt2 = e2->get_result_type();
  suif_assert(dt1);
  suif_assert(dt2);
  if (dt1 != dt2 || !(is_a<IntegerType>(dt1) || is_a<FloatingPointType>(dt1))){
    report_error(token, "arguments of `%s' must be numeric and match", op.c_str());
    report_warning(token, "left argument type is `%s'", dt1->get_name().c_str());
    report_warning(token, "right argument type is `%s'", dt2->get_name().c_str());
  }
  return e;
}

Expression *EsterelTreeParser::build_boolean_unary(RefLineAST token,
						   Expression *e1)
{
  suif_assert(e1);
  LString op(token->getText().c_str());
  Expression *e = create_unary_expression(suif_env, booleanType, op, e1);

  DataType *dt1 = e1->get_result_type();
  suif_assert(dt1);
  if (!is_a<BooleanType>(dt1)) {
    report_error(token, "arguments of `%s' must be boolean", op.c_str());
    report_warning(token, "argument type is `%s'", dt1->get_name().c_str());
  }
  return e;
}

Expression *EsterelTreeParser::build_numeric_unary(RefLineAST token,
						   Expression *e1)
{
  suif_assert(e1);
  LString op(token->getText().c_str());
  Expression *e =
    create_unary_expression(suif_env, e1->get_result_type(), op, e1);

  DataType *dt1 = e1->get_result_type();
  suif_assert(dt1);
  if (is_a<BooleanType>(dt1)) {
    report_error(token, "arguments of `%s' must be boolean", op.c_str());
    report_warning(token, "argument type is `%s'", dt1->get_name().c_str());
  }
  return e;
}

void EsterelTreeParser::create_constant(Context *c,
					RefLineAST id,
					Expression *e,
					RefLineAST expr,
					QualifiedType *qt)
{
  LString name(id->getText().c_str());
  SymbolTableObject *sto = c->variables->lookup_lookup_table(name);
  if (sto) {
    report_error(id, "redeclaration of constant `%s'", name.c_str());
    report_warning(sto, "`%s' previously declared here", name.c_str());
  }

  QualifiedType *qtc = get_const_type(c, qt->get_base_type());

  // Add a VariableSymbol to md->symbol_table
  //   QualifiedType must be a variant with const
  //   VariableDefinition
  //
  // Add VariableDefinition to md->definition_block->variable_definitions
  //   VariableSymbol points to above
  //   ValueBlock is an ExpressionValueBlock      

  VariableSymbol *vs = create_variable_symbol(suif_env, qtc);
  attach_token_linenumber(vs, id);
  c->variables->add_symbol(name, vs);

  ValueBlock *vb = e ? create_expression_value_block(suif_env, e) : 0;
  VariableDefinition *vd = create_variable_definition(suif_env, vs, 0, vb);
  c->module->get_definition_block()->append_variable_definition(vd);

  vs->set_definition(vd);

  if (e) {
    if (e->get_result_type() != qt->get_base_type()) {
      report_error(id, "initializer does not match type of its constant");
      report_warning(id, "type of constant is `%s'", qt->get_name().c_str());
      report_warning(expr, "type of initializer is `%s'",
		     e->get_result_type()->get_name().c_str());
    }
  }
}

void EsterelTreeParser::create_local_variable(Context *c,
					      VariableStatement *vst,
					      RefLineAST id,
					      Expression *e,
					      RefLineAST expr,
					      QualifiedType *qt)
{
  LString name(id->getText().c_str());
  suif_assert(c->variables);
  SymbolTableObject *sto = c->variables->lookup_lookup_table(name);
  if (sto) {
    report_error(id, "redefinition of local variable `%s'",
		 name.c_str());
    report_warning(sto, "`%s' previously defined here", name.c_str());
  }
  VariableSymbol *vsy = create_variable_symbol(suif_env, qt);
  attach_token_linenumber(vsy, id);
  c->variables->add_symbol(name, vsy);

  ValueBlock *vb = e ? create_expression_value_block(suif_env, e) : 0;
  VariableDefinition *vd =
    create_variable_definition(suif_env, vsy, 0, vb);
  vst->get_definition_block()->append_variable_definition(vd);
  vsy->set_definition(vd);

  if (e) {
    if (e->get_result_type() != qt->get_base_type()) {
      report_error(id, "initializer does not match type of its variable");
      report_warning(id, "type of variable is `%s'", qt->get_name().c_str());
      report_warning(expr, "type of initializer is `%s'",
		     e->get_result_type()->get_name().c_str());
    }
  }
}

void EsterelTreeParser::attach_token_linenumber(AnnotableObject *ao,
						RefLineAST rt)
{
  IntegerBrick *ib = create_integer_brick(suif_env,rt->getLine());
  StringBrick *sb = create_string_brick(suif_env,current_filename);
  BrickAnnote *ba = create_brick_annote(suif_env,lineAnnote);
  ba->append_brick(ib);
  ba->append_brick(sb);
  ao->append_annote(ba);
}

void EsterelTreeParser::report_warning( AnnotableObject* ao,
					const char* format, ... )
{
  report_fileline(ao);
  va_list args;
  va_start( args, format );
  vfprintf( stderr, format, args );
  fprintf( stderr, "\n" );
}


void EsterelTreeParser::report_error( AnnotableObject* ao,
				      const char* format, ... )
{
  report_fileline(ao);
  va_list args;
  va_start( args, format );
  vfprintf( stderr, format, args );
  fprintf( stderr, "\n" );
  num_errors++;
}

void EsterelTreeParser::report_error( RefLineAST ast,
				      const char* format, ... )
{
  fprintf( stderr,  "%s:%d: ", current_filename.c_str(),
	                             ast->getLine() );
  va_list args;
  va_start( args, format );
  vfprintf( stderr, format, args );
  fprintf( stderr, "\n" );
  num_errors++;
}

void EsterelTreeParser::report_warning( RefLineAST ast,
				      const char* format, ... )
{
  fprintf( stderr,  "%s:%d: ", current_filename.c_str(),
	                             ast->getLine() );
  va_list args;
  va_start( args, format );
  vfprintf( stderr, format, args );
  fprintf( stderr, "\n" );
}

void EsterelTreeParser::report_fileline( AnnotableObject* ao )
{
  Annote * line = ao->lookup_annote_by_name(lineAnnote);
  if (line) {
    BrickAnnote *ba = to<BrickAnnote>(line);
    suif_assert( ba->get_brick_count() == 2 );
    IntegerBrick *ib = to<IntegerBrick>(ba->get_brick(0));
    StringBrick *sb = to<StringBrick>(ba->get_brick(1));
    fprintf( stderr, "%s:%d: ", sb->get_value().c_str(),
                                ib->get_value().c_int() );
  }
}

}

options {
  language="Cpp";
}

/**
 *
 * Tree walker for Esterel AST that builds the SUIF representation
 *
 * Author: Stephen A. Edwards
 *         sedwards@cs.columbia.edu
 *
 *
 * Change Log
 */
class EsterelTreeParser extends TreeParser;

options {
  importVocab = Esterel;
  ASTLabelType = "RefLineAST";	// Expect AST nodes with line numbers
  //  defaultErrorHandler=false;	// Disable default error handling for
}

{
  SuifEnv *suif_env;
  LString current_filename;
  int num_errors;

  static LString inputAnnote;
  static LString outputAnnote;
  static LString inputoutputAnnote;
  static LString returnAnnote;
  static LString sensorAnnote;
  static LString localAnnote;
  static LString integerTypeName;
  static LString booleanTypeName;
  static LString stringTypeName;
  static LString floatTypeName;
  static LString doubleTypeName;
  static LString lineAnnote;
  static LString constantQualifier;
  static LString tickSignalName;
  static LString voidTypeName;

  struct Context {
    ModuleDefinition *module;
    SymbolTable *types;
    SymbolTable *functions;
    SymbolTable *procedures;
    SymbolTable *tasks;
    SymbolTable *variables;
    SymbolTable *traps;
    SymbolTable *signals;    
  };

public:
  void set_suif_env(SuifEnv *env) { suif_env = env; }

  /// Built-in types
  IntegerType *integerType;
  QualifiedType *qIntegerType;
  BooleanType *booleanType;
  StringType *stringType;
  FloatingPointType *floatType;
  FloatingPointType *doubleType;
  VoidType *voidType;
  QualifiedType *qVoidType;

  /// Find or create a built-in function with the given name
  BuiltinFunctionSymbol *get_builtin_function(Context *, LString);

  /// Find or create a constant type with the given base type
  QualifiedType *get_const_type(Context *, DataType *);

  /// Locate a given variable symbol or produce an error
  VariableSymbol *find_variable(Context *, RefLineAST);

  /// Locate a given exception symbol or produce an error
  TrapSymbol *find_trap(Context *, RefLineAST);

  /// Locate a given signal symbol or produce an error
  SignalSymbol *find_signal(Context *, RefLineAST);

  /// Construct and typecheck a Boolean (and, or) binary operator
  Expression *build_boolean_binary(RefLineAST, Expression*, Expression*);

  /// Construct and typecheck a Numeric (+, *) binary operator
  Expression *build_numeric_binary(RefLineAST, Expression*, Expression*);

  /// Construct and typecheck a equality (=, <>) binary operator
  Expression *build_equality_binary(RefLineAST, Expression*, Expression*);

  /// Construct and typecheck a comparison (<, <=, etc.) binary operator
  Expression *build_comparison_binary(RefLineAST, Expression*, Expression*);

  /// Construct and typecheck a boolean (not) unary operator
  Expression *build_boolean_unary(RefLineAST, Expression*);

  /// Construct and typecheck a numeric (-) unary operator
  Expression *build_numeric_unary(RefLineAST, Expression*);

  /// Create a constant in the module
  void create_constant(Context *, RefLineAST, Expression *,
		       RefLineAST, QualifiedType *);

  /// Create a local variable in a variable statement
  void create_local_variable(Context *, VariableStatement *,  RefLineAST,
			     Expression *, RefLineAST, QualifiedType *);

  /// Attach line number information to an AnnotableObject
  void attach_token_linenumber(AnnotableObject *, RefLineAST);


  /// Report a warning
  void report_warning(AnnotableObject *, const char *, ... );
  void report_warning(RefLineAST, const char *, ... );

  /// Report an error
  void report_error(AnnotableObject *, const char *, ... );
  void report_error(RefLineAST, const char *, ... );

  /// Return the number of errors
  int get_num_errors() const { return num_errors; }

  /// Report file:line: for an annotated object.
  void report_fileline(AnnotableObject*);
  
}

file [FileSetBlock* fsb, LString filename]
  {
    FileBlock *fb = create_file_block(suif_env, filename);
    fsb->append_file_block(fb);
    current_filename = filename;
    num_errors = 0;
  }
  : (module[fb])+
  ;
/*
exception
catch [ANTLR_USE_NAMESPACE(antlr)NoViableAltException& ex] {
  report_error(static_cast<RefLineAST>(ex.node),
	       "Unexpected AST node `%s'",
	       ex.toString().c_str());
}
*/

/*
 * Line number information:
 * A BrickAnnote called "line" contains an IntegerBrick with the line number
 * and a StringBrick with the filename.
 */

module [FileBlock* fb]
  : #( "module" moduleName:ID
       {
	 // Check for a duplicate module definition

	 LString name(moduleName->getText().c_str());
         SymbolTableObject* sto =
           fb->get_symbol_table()->lookup_lookup_table(name);
         if ( sto ) {	   
	   report_error(moduleName, "redefinition of module `%s'",
                                      name.c_str() );
           report_warning(sto, "`%s' previously defined here", name.c_str());
         }

         ModuleSymbol *ms = create_module_symbol(suif_env, (ProcedureType*) 0);
	 attach_token_linenumber(ms, moduleName);
         fb->get_symbol_table()->add_symbol(name, ms);

         // Create a new module and add it to the FileBlock

         ModuleDefinition *md = create_module_definition(suif_env);
         md->set_procedure_symbol(ms);
	 attach_token_linenumber(md, moduleName);
         fb->get_definition_block()->append_procedure_definition(md);

	 Context c;
	 c.module = md;
	 c.types = md->get_type_symbols();
	 c.functions = md->get_function_symbols();
	 c.procedures = md->get_procedure_symbols();
	 c.tasks = md->get_task_symbols();
	 c.variables = md->get_symbol_table();
	 c.traps = 0;	// No traps at the outer level
	 c.signals = md->get_signal_symbols();

         // Add the built-in types

	 QualifiedType *qt;
	 integerType =
	   create_integer_type(suif_env,32,32,true,integerTypeName);
         c.types->append_symbol_table_object(integerType);
         qIntegerType = create_qualified_type(suif_env, integerType);
         c.types->add_symbol(integerTypeName, qIntegerType);

         booleanType = create_boolean_type(suif_env,32,32,booleanTypeName);
         c.types->append_symbol_table_object(booleanType);
         qt = create_qualified_type(suif_env, booleanType);
         c.types->add_symbol(booleanTypeName, qt);

         stringType = create_string_type(suif_env,32,32,stringTypeName);
         c.types->append_symbol_table_object(stringType);
         qt = create_qualified_type(suif_env, stringType);
         c.types->add_symbol(stringTypeName, qt);

         floatType = create_floating_point_type(suif_env,32,32,floatTypeName);
         c.types->append_symbol_table_object(floatType);
         qt = create_qualified_type(suif_env, floatType);
         c.types->add_symbol(floatTypeName, qt);

         doubleType =
	   create_floating_point_type(suif_env,64,64,doubleTypeName);
         c.types->append_symbol_table_object(doubleType);
         qt = create_qualified_type(suif_env, doubleType);
         c.types->add_symbol(doubleTypeName, qt);

	 voidType = create_void_type(suif_env,0,0, voidTypeName);
	 c.types->append_symbol_table_object(voidType);
	 qVoidType = create_qualified_type(suif_env, voidType, voidTypeName);
         c.types->append_symbol_table_object(qVoidType);

	 // Built-in input signal "tick"
	 SignalType *st =
	   create_signal_type(suif_env, voidType, inputAnnote, voidTypeName);
	 c.module->get_signal_symbols()->append_symbol_table_object(st);
	 SignalSymbol *ss = create_signal_symbol(suif_env, st);
	 c.signals->add_symbol(tickSignalName,ss);
	 VariableDefinition *vd =
	   create_variable_definition(suif_env, ss, 0, 0);
	 ss->set_definition(vd);
	 c.module->get_definition_block()->append_variable_definition(vd);
       }
       declarations[&c]
       {
         Statement *sta;
       }
       sta=statement[&c]
       {
         md->set_body(sta); 
       }
     )
  ;

declarations [Context *c]
  : #( DECLS
       ( #( "type" ( typeDecl[c] )+ )
       | #( "constant" (constantDecl[c])+ )
       | #( "input" ( signalDecl[c, inputAnnote] )+ )
       | #( "output" ( signalDecl[c, outputAnnote] )+ )
       | #( "inputoutput" ( signalDecl[c, inputoutputAnnote] )+ )
       | #( "return" ( signalDecl[c, returnAnnote] )+ )
       | #( "function" ( functionDecl[c] )+ )
       | #( "procedure" ( procedureDecl[c] )+ )
       | #( "task" ( taskDecl[c] )+ )
       | #( "sensor" ( sensorDecl[c] )+ )
       | #( "relation" ( relationDecl[c] )+ )
       )*
     )
  ;

typeDecl [Context *c]
  : typeName:ID {
       LString name(typeName->getText().c_str());
       SymbolTableObject *sto =
          c->types->lookup_lookup_table(name);
       if (sto) {
         if ( is_a<ExternalType>(to<QualifiedType>(sto)->get_base_type()) ) {
           report_error(typeName, "redeclaration of type `%s'", name.c_str());
           report_warning(sto, "`%s' previously declared here", name.c_str());
	} else {
           report_error(typeName, "redeclaration of built-in type `%s'", name.c_str());
        }
       }
       ExternalType *et = create_external_type(suif_env,0,0,name);
       c->types->append_symbol_table_object(et);
       QualifiedType *qt = create_qualified_type(suif_env, et);
       attach_token_linenumber(qt, typeName);
       c->types->add_symbol(name, qt);
    }
  ;

constantDecl [Context *c]
  { Expression *e = 0;
    QualifiedType *qt; }
  : #(CDECL
      ( constantName:ID ( #(EQUALS e=expr:expression[c]) )?
        qt=typeToken:type[c]
        { create_constant(c, constantName, e, expr, qt); }
      | #(ids:COMMA (ID)+) qt=type[c]
        {
	  for ( RefLineAST id = #ids->getFirstChild() ; id ;
		id=id->getNextSibling() )
	    create_constant(c, id, 0, 0, qt);
	}
      )
    )
  ;

signalDecl [Context *c, LString direction]
  : #(SDECL signalName:ID
       {
         LString name(signalName->getText().c_str());
         SymbolTableObject *sto =
            c->signals->lookup_lookup_table(name);

         if (sto) {
           report_error(signalName, "redeclaration of signal `%s'", name.c_str());
           report_warning(sto, "`%s' previously declared here", name.c_str());
         }
         Expression *e = 0;
         SignalType *st;
       }
       ( #(COLEQUALS e=expr:expression[c]) )?
       st=typeToken:channelType[c,direction]
       {
         SignalSymbol *ss = create_signal_symbol(suif_env, st);
         attach_token_linenumber(ss, signalName);
         c->signals->add_symbol(name,ss);

         ValueBlock *vb = e ? create_expression_value_block(suif_env, e) : 0;
	 VariableDefinition *vd =
	   create_variable_definition(suif_env, ss, 0, vb);
	 c->module->get_definition_block()->append_variable_definition(vd);

	 ss->set_definition(vd);

	 if (e) {
	   suif_assert(st->get_base_type());
	   if (e->get_result_type() != st->get_base_type()) {
	     report_error(typeToken,
			  "initializer does not match type of signal");
	     report_warning(typeToken, "type of signal is `%s'",
			    st->get_name().c_str());
	     report_warning(expr, "type of initializer is `%s'",
			    e->get_result_type()->get_name().c_str());
	   }
	 }
       }
     )
  ;

type [Context *c] returns [QualifiedType* qt]
  :
    type:ID {
      LString name(type->getText().c_str());
      SymbolTableObject *sto = c->types->lookup_lookup_table(name);
      if (!sto) {
        report_error(type, "unknown type `%s'", name.c_str());
        // Define the type to avoid future errors
	ExternalType *et = create_external_type(suif_env,0,0,name);
	c->types->append_symbol_table_object(et);
        qt = create_qualified_type(suif_env, et);
        attach_token_linenumber(et, type);
        c->types->add_symbol(name, qt);
      } else {
        qt = to<QualifiedType>(sto);
      }
    }
  ;

// FIXME: Typecheck the combine function
channelType [Context *c, LString direction] returns [SignalType *st]
  {
    QualifiedType *qt = qVoidType;
    EsterelFunctionSymbol *efs = 0;
  }
  : (qt=typeToken:type[c]
      ( func:ID
        {
	  LString name = func->getText().c_str();
	  SymbolTableObject *sto = c->functions->lookup_lookup_table(name);
	  if (!sto)
	    report_error(func, "undeclared function `%s'", name.c_str());
	  else
	    efs = to<EsterelFunctionSymbol>(sto);
	}
      | pcf:predefinedCombineFunction
        { efs = get_builtin_function(c, pcf->getText().c_str()); }
      )?
    )?
    {
      DataType *dt = qt->get_base_type();
      Iter<SymbolTableObject *> iter =
	c->module->get_signal_symbols()->get_symbol_table_object_iterator();
      while (iter.is_valid()) {
	SymbolTableObject *sto = iter.current();
	if ( is_a<SignalType>(sto) ) {
	  st = to<SignalType>(sto);
	  if ( st->get_base_type() == dt &&
	       st->get_direction() == direction &&
	       st->get_combine() == efs )
	    goto GotType;
	}
	iter.next();
      }
      st = create_signal_type(suif_env, dt, direction, dt->get_name(), efs);
      c->module->get_signal_symbols()->append_symbol_table_object(st);

    GotType:
      suif_assert(st);
      suif_assert(st->get_base_type());
    }
  ;

predefinedCombineFunction
  : PLUS
  | STAR
  | "or"
  | "and"
  ;

functionDecl [Context *c]
  : #( FDECL functionName:ID
    {
      LString name(functionName->getText().c_str());
      SymbolTableObject *sto =
        c->functions->lookup_lookup_table(name);
      if (sto) {
        report_error(functionName,
		     "redeclaration of function `%s'", name.c_str());
        report_warning(sto, "`%s' previously declared here", name.c_str());
      }
      EsterelFunctionType *eft = create_esterel_function_type(suif_env);
      c->functions->append_symbol_table_object(eft);
      EsterelFunctionSymbol *efs =
        create_esterel_function_symbol(suif_env, eft);
      attach_token_linenumber(efs, functionName);
      c->functions->add_symbol(name,efs);
      QualifiedType *qt, *rt;
    }
    #(TYPES (qt=type[c] { eft->append_argument(qt); } )* )
    rt=type[c]
    {
      eft->set_result(rt);
    }
    )
  ;

procedureDecl [Context *c]
  : #( PDECL procedureName:ID
    {
      LString name(procedureName->getText().c_str());
      SymbolTableObject *sto =
        c->procedures->lookup_lookup_table(name);
      if (sto) {
        report_error(procedureName, "redeclaration of procedure `%s'", name.c_str());
        report_warning(sto, "`%s' previously declared here", name.c_str());
      }
      EsterelProcedureType *ept = create_esterel_procedure_type(suif_env);
      c->procedures->append_symbol_table_object(ept);
      EsterelProcedureSymbol *eps =
         create_esterel_procedure_symbol(suif_env, ept);
      attach_token_linenumber(eps, procedureName);
      c->procedures->add_symbol(name,eps);
      QualifiedType *qt;
    }
    #(TYPES (qt=type[c] { ept->append_reference_argument(qt); } )* )
    #(TYPES (qt=type[c] { ept->append_value_argument(qt); } )* )  
    )
  ;

taskDecl [Context *c]
  : #( TADECL taskName:ID
    {
      LString name(taskName->getText().c_str());
      SymbolTableObject *sto =
        c->tasks->lookup_lookup_table(name);
      if (sto) {
        report_error(taskName, "redeclaration of task `%s'", name.c_str());
        report_warning(sto, "`%s' previously declared here", name.c_str());
      }
      EsterelProcedureType *ept = create_esterel_procedure_type(suif_env);
      c->tasks->append_symbol_table_object(ept);
      EsterelProcedureSymbol *eps =
         create_esterel_procedure_symbol(suif_env, ept);
      attach_token_linenumber(eps, taskName);
      c->tasks->add_symbol(name,eps);
      QualifiedType *qt;
    }
    #(TYPES (qt=type[c] { ept->append_reference_argument(qt); } )* )
    #(TYPES (qt=type[c] { ept->append_value_argument(qt); } )* )  
    )
  ;

sensorDecl [Context *c]
  : #(SEDECL sensorName:ID
       {
         LString name(sensorName->getText().c_str());
         SymbolTableObject *sto =
            c->signals->lookup_lookup_table(name);

         if (sto) {
           report_error(sensorName, "redeclaration of sensor `%s'", name.c_str());
           report_warning(sto, "`%s' previously declared here", name.c_str());
         }

         SignalType *st;
       }
       st=channelType[c, sensorAnnote]
       {
         SignalSymbol *ss = create_signal_symbol(suif_env, st);
         attach_token_linenumber(ss, sensorName);
         c->signals->add_symbol(name,ss);

	 VariableDefinition *vd =
	   create_variable_definition(suif_env, ss, 0, 0);
	 c->module->get_definition_block()->append_variable_definition(vd);
	 ss->set_definition(vd);
       }
     )
  ;

relationDecl [Context *c]
  : #( IMPLIES sig1:ID sig2:ID
      {
	SignalSymbol *ss1 = find_signal(c, sig1);
	SignalSymbol *ss2 = find_signal(c, sig2);
	ImpliesRelation *ir = create_implies_relation(suif_env, ss1, ss2);
	attach_token_linenumber(ir, #relationDecl);
	c->module->append_relation(ir);
      }
     )
  | #( POUND
      {
	ExclusionRelation *er = create_exclusion_relation(suif_env);
	attach_token_linenumber(er, #relationDecl);
	c->module->append_relation(er);
      }
      (sig:ID
        {
	  SignalSymbol *ss = find_signal(c, sig);
	  er->append_signal(ss);
	}
      )+
     )    
  ; 



// ModuleDefinition contains the symbol table for functions, procedures, tasks.
// vs, ts, and ss are the innermost scopes for variables/constants,
// traps, and signals

statement [Context *c] returns [Statement* st]
  : ( #( PARALLEL { ParallelStatementList *sl =
                            create_parallel_statement_list(suif_env); }
         (st=statement[c] { sl->append_statement(st); } )+
       ) { st = sl; }

    | #( SEQUENCE { StatementList *sl = create_statement_list(suif_env); }
         (st=statement[c] { sl->append_statement(st); } )+
       ) { st = sl; }

    | "nothing" { st = create_nothing_statement(suif_env); }

    | "pause" { st = create_pause_statement(suif_env); }

    | "halt" { st = create_halt_statement(suif_env); }

    | #( "emit"
	 esig:ID
         {
           SignalSymbol *ss = find_signal(c, esig);
           Expression * e = 0;
         }
         ( e=expression[c] )?
         {
           EmitStatement *es = create_emit_statement( suif_env, ss, e );
           if ( e && ss->get_type()->get_base_type() == voidType ) {
	     report_error(esig, "illegal emission of a value on pure signal `%s'",
			  esig->getText().c_str());
	   }
           st = es;
         }
      )

    | #( "sustain"
	 ssig:ID
         {
           SignalSymbol *ss = find_signal(c, ssig);
           Expression * e = 0;
         }
         ( e=expression[c] )?
         {
           EmitStatement *es = create_sustain_statement( suif_env, ss, e );
           if ( e && ss->get_type()->get_base_type() == voidType ) {
	     report_error(ssig, "illegal sustain of a value on pure signal `%s'",
			  ssig->getText().c_str());
	   }
           st = es;
         }
      )

    | { Expression *e; }
      #( COLEQUALS var:ID e=expression[c] )
      { 
        VariableSymbol *vs = find_variable(c, var);
        st = create_store_variable_statement(suif_env, vs, e);
        DataType *dt1 = vs->get_type()->get_base_type();
        DataType *dt2 = e->get_result_type();
        if ( dt1 != dt2 ) {
          report_error(#statement,
	   "type mismatch: expression type `%s' must match variable type `%s'",
           dt2->get_name().c_str(), dt1->get_name().c_str());
        }
      }

    | #( "call" proc:ID
         {
           LString name = proc->getText().c_str();
	   suif_assert(c->procedures);
	   SymbolTableObject *sto = c->procedures->lookup_lookup_table(name);
	   if (!sto) {
	     report_error(proc, "undeclared procedure `%s'", name.c_str());
	     // Create a nothing statement placeholder
	     st = create_nothing_statement(suif_env);
	     break;
	   }
	   suif_assert(sto);
	   EsterelProcedureSymbol *eps = to<EsterelProcedureSymbol>(sto);
	   suif_assert(eps);
	   EsterelProcedureType *ept = eps->get_type();
	   suif_assert(ept);
	   ProcedureCallStatement *pcs =
	     create_procedure_call_statement(suif_env, eps);
         }
	 procArgs[c, pcs]
         { st = pcs; }
      )

    | #("present"
      {
        PresentStatement *ps = create_present_statement(suif_env);      
      }
      ( #(ca:CASE
          { Expression *e; Statement *s = 0; }
          e=sigExpression[c] (s=statement[c])?
          {
	    PredicatedStatement *ps1 =
	      create_predicated_statement(suif_env, s, e);
            attach_token_linenumber(ps1, ca);
            ps->append_case(ps1);
          }
        ) )*     
      ( #("else"
          { Statement *s = 0; }
          s=statement[c]
          { ps->set_default_case(s); }
        )
      )?
      { st = ps; }
      )
             
    | #( "if"
        {
          EsterelIfStatement *is = create_esterel_if_statement(suif_env);
	  Expression *e;
          PredicatedStatement *ps;
        }
        e=ife:expression[c]
        { Statement *s = 0; }
        ( #("then" s=statement[c] ) )?
 	{
	  ps = create_predicated_statement(suif_env, s, e); 
	  attach_token_linenumber(ps, ife);
	  is->append_case(ps);
	}
	( #(elif:"elsif" e=expression[c] s=statement[c] )
	    {
	      ps = create_predicated_statement(suif_env, s, e);
	      attach_token_linenumber(ps, elif);
	      is->append_case(ps);
	    }
	)*
        ( #("else" s=statement[c] ) { is->set_default_case(s); } )?
        { st = is; }
      )

    | #( "loop"
        { Statement *s; Expression *e; }
        s=statement[c]
        (
           /* nothing */ { st = create_loop_statement(suif_env, s); }
	 | e=delayExpression[c]
  	   { st = create_loop_each_statement(suif_env, s, e); }
        )
      )

    | #("repeat"
        {
	  bool positive = false;
          Expression *e;
        }
        ("positive" { positive=true; } )?
        e=expr:expression[c]
        {
          if ( e->get_result_type() != integerType ) {
	    report_error( expr,
			  "type of repeat count is `%s' expecting integer",
			  e->get_result_type()->get_name().c_str());
          }
	  Statement *s;
        }
        s=statement[c]
        {
	  st = create_repeat_statement(suif_env, s, e, positive); 
        }       
      )

    | #("abort"
        { bool isWeak; Statement *s; }
        ("weak" { isWeak = true; })?
        s=statement[c]
        {
	  AbortStatement *as = create_abort_statement(suif_env, s, 0, isWeak);
	}
        ( #(ca1:CASE { Expression *e; s = 0; }
            e=delayExpression[c] (s=statement[c])?
            {
              PredicatedStatement *ps =
		create_predicated_statement(suif_env, s, e);
	      attach_token_linenumber(ps, ca1);
	      as->append_case(ps);
	    }
	) )+
        { st = as; }
      )

    | #("await"
        { AwaitStatement *as = create_await_statement(suif_env); }
        ( #(ca2:CASE { Expression *e; Statement *s = 0; }
            e=delayExpression[c] (s=statement[c])?
            {
              PredicatedStatement *ps =
		create_predicated_statement(suif_env, s, e);
	      attach_token_linenumber(ps, ca2);
	      as->append_case(ps);
	    }
	) )+
        { st = as; }
      )

    | #("every" { Expression *e; Statement *s; }
        e=delayExpression[c] s=statement[c]
        { st = create_every_statement(suif_env, s, e); }
      )

    | #("suspend" { Expression *e; Statement *s; }
         s=statement[c] e=delayExpression[c]
        { st = create_suspend_statement(suif_env, s, e); }
      )

    | #( "trap"
      {
         TrapStatement *ts = create_trap_statement(suif_env);
	 Context nc = *c;
	 nc.traps = ts->get_symbol_table();
	 nc.traps->set_explicit_super_scope(c->traps);
      }
      #(TRAPS ( #(TDECL trapName:ID
		{ QualifiedType *qt = qVoidType; Expression *e = 0; }
		( (#(COLEQUALS e=expression[c]) )? qt=typeToken:type[c] )?
		{
		  LString name(trapName->getText().c_str());
		  SymbolTableObject *sto = nc.traps->lookup_lookup_table(name);
		  if (sto) {
		    report_error(trapName, "redefinition of trap `%s'",
				 name.c_str());
		    report_warning(sto, "`%s' previously defined here",
				   name.c_str());
		  }
		  TrapSymbol *tsy = create_trap_symbol(suif_env, qt);
		  attach_token_linenumber(tsy, trapName);
		  nc.traps->add_symbol(name, tsy);

		  ValueBlock *vb =
		    e ? create_expression_value_block(suif_env, e) : 0;
		  VariableDefinition *vd =
		    create_variable_definition(suif_env, tsy, 0, vb);
		  ts->get_definition_block()->append_variable_definition(vd);
		  tsy->set_definition(vd);

		  if (e && e->get_result_type() != qt->get_base_type() ) {
		    report_error(trapName,
				 "initializer does not match type of trap");
		    report_warning(typeToken, "type of trap is `%s'",
				   qt->get_name().c_str());
		    report_warning(e, "type of initializer is `%s'",
				   e->get_result_type()->get_name().c_str());
		  }
		}
	      )
           )+
     )
     { Statement *s; }
     s=statement[&nc]
     {
       ts->set_body(s);
       Expression *e;
     }
     ( #(htok:"handle" e=trapExpression[&nc] s=statement[&nc]
	 {
	   PredicatedStatement *ps =
	     create_predicated_statement(suif_env, s, e);
	   attach_token_linenumber(ps, htok);
	   ts->append_handler(ps);
         }
	 ))*
     { st = ts; }
     )

    | #( "exit" etrap:ID { Expression *e = 0; } ( e=texpr:expression[c] )?
      {
        TrapSymbol *ts = find_trap(c, etrap);
	st = create_exit_statement( suif_env, ts, e);
	if (e) {
	  if ( ts->get_type() == qVoidType ) {
	    report_error(etrap, "attempting to return value with pure trap `%s'",
			 ts->get_name().c_str());
	  } else if ( e->get_result_type() !=
		      ts->get_type()->get_base_type() ) {
	    report_error(etrap,
			 "returning wrong type for trap `%s'",
			 ts->get_name().c_str());
	    report_warning(ts, "trap of type %s",
			   ts->get_type()->get_name().c_str());
	    report_warning(texpr, "expression of type %s",
			   e->get_result_type()->get_name().c_str());
	  }
	} else {
	  // No expression
	  if ( ts->get_type() != qVoidType ) {
	    report_error(etrap,
			 "valued trap `%s' requires exit with an expression",
			 ts->get_name().c_str());
	    report_warning(ts, "trap has type `%s'",
			   ts->get_type()->get_name().c_str());
	  }
	}
      }
      )

    | #("exec"
        { ExecStatement *es = create_exec_statement(suif_env); }
	(taskCall[c, es])+
        { st = es; }
       )

    | #( "var" {
         VariableStatement *vst = create_variable_statement(suif_env);
	 Context nc = *c;
	 nc.variables = vst->get_symbol_table();
	 nc.variables->set_explicit_super_scope(c->variables);	 
         QualifiedType *qt;
      }
      #(VARS
        ( #(VDECL
            ( variableName:ID
	      { Expression *e = 0; }
	      ( #(COLEQUALS e=varexpr:expression[c]) )?
	      qt=type[c]
	      { create_local_variable(&nc,vst,variableName,e,varexpr,qt); }
	    | #(ids:COMMA (ID)+) qt=type[c]
	        {
		  for ( RefLineAST id = #ids->getFirstChild() ; id ;
			id = id->getNextSibling() )
		    create_local_variable(&nc,vst,id,0,0,qt);
		}	
            )
          )
        )+
       )
       { Statement *s; }
       s=statement[&nc]
       {
	 vst->set_body(s);
	 st = vst;
       }
      )

    | #( "signal"
        {
	  SignalStatement *ss = create_signal_statement(suif_env);
	  Context nc = *c;
	  nc.signals = ss->get_symbol_table();
	  nc.signals->set_explicit_super_scope(c->signals);
        }
      #(SIGS ( #(SDECL sigName:ID
		{ SignalType *st; Expression *e = 0; }
		( #(COLEQUALS e=expression[c]) )?
		st=stypeToken:channelType[c,localAnnote]
		{
		  LString name(sigName->getText().c_str());
		  SymbolTableObject *sto =
		    nc.signals->lookup_lookup_table(name);
		  if (sto) {
		    report_error(sigName, "redefinition of signal `%s'",
				 name.c_str());
		    report_warning(sto, "`%s' previously defined here",
				   name.c_str());
		  }

		  SignalSymbol *sig = create_signal_symbol(suif_env, st);
		  attach_token_linenumber(sig, sigName);
		  nc.signals->add_symbol(name, sig);

		  ValueBlock *vb =
		    e ? create_expression_value_block(suif_env, e) : 0;
		  VariableDefinition *vd =
		    create_variable_definition(suif_env, sig, 0, vb);
		  ss->get_definition_block()->append_variable_definition(vd);
		  sig->set_definition(vd);

		  if (e && e->get_result_type() != st->get_base_type() ) {
		    report_error(sigName,
				 "initializer does not match type of signal");
		    report_warning(stypeToken, "type of signal is `%s'",
				   st->get_name().c_str());
		    report_warning(e, "type of initializer is `%s'",
				   e->get_result_type()->get_name().c_str());
		  }
		}
	      )
           )+
     )
     { Statement *s; }
     s=statement[&nc]
     {
       ss->set_body(s);
       st = ss;
     }
    )

    | #(RUN module:ID
        {
	  RunStatement *rs =
	    create_run_statement(suif_env, module->getText().c_str(),
				 module->getText().c_str());
	}
        ( oldMod:ID { rs->set_old_module_name(oldMod->getText().c_str()); } )?
        ( typeRenaming[c,rs]
        | constantRenaming[c,rs]
        | functionRenaming[c,rs]
        | procedureRenaming[c,rs]
        | taskRenaming[c,rs]
        | signalRenaming[c,rs]
        )*
	{ st = rs; }
      )

    | #(DOWATCHING { Expression *e; Statement *s1, *s2 = 0; }
        s1=statement[c]
        e=de:delayExpression[c]
        (s2=statement[c])?
        { st = create_do_watching_statement(suif_env, s1, e, s2); }
      )
    | #(DOUPTO { Expression *e; Statement *s; }
        s=statement[c]
        e=delayExpression[c]
        { st = create_do_upto_statement(suif_env, s, e); }
    )
    { attach_token_linenumber(st,#statement); }
    )
  ;


expression [Context *c] returns [Expression *e]
  : { Expression *e1, *e2; }
    (
      #("or" e1=expression[c] e2=expression[c] )
      { e = build_boolean_binary( #expression, e1, e2 ); }
    | #("and" e1=expression[c] e2=expression[c] )
      { e = build_boolean_binary( #expression, e1, e2 ); }

    | #(PLUS e1=expression[c] e2=expression[c] )
      { e = build_numeric_binary( #expression, e1, e2 ); }
    | #(DASH e1=expression[c] (
           e2=expression[c]
	   { e = build_numeric_binary( #expression, e1, e2 ); }
         | /* nothing */ { e = build_numeric_unary( #expression, e1 ); }
         ))

    | #(STAR e1=expression[c] e2=expression[c] )
      { e = build_numeric_binary( #expression, e1, e2 ); }
    | #(SLASH e1=expression[c] e2=expression[c] )
      { e = build_numeric_binary( #expression, e1, e2 ); }
    | #("mod" e1=expression[c] e2=expression[c] )
      { e = build_numeric_binary( #expression, e1, e2 ); }

    | #(EQUALS e1=expression[c] e2=expression[c] )
      { e = build_equality_binary( #expression, e1, e2 ); }
    | #(NEQUAL e1=expression[c] e2=expression[c] )
      { e = build_equality_binary( #expression, e1, e2 ); }

    | #(LESSTHAN e1=expression[c] e2=expression[c] )
      { e = build_comparison_binary( #expression, e1, e2 ); }
    | #(GREATERTHAN e1=expression[c] e2=expression[c] )
      { e = build_comparison_binary( #expression, e1, e2 ); }
    | #(LEQUAL e1=expression[c] e2=expression[c] )
      { e = build_comparison_binary( #expression, e1, e2 ); }
    | #(GEQUAL e1=expression[c] e2=expression[c] )
      { e = build_comparison_binary( #expression, e1, e2 ); }

    | #("not" e1=expression[c])
      { e = build_boolean_unary( #expression, e1 ); }

    | Integer {
        e = create_int_constant( suif_env, integerType,
				 #expression->getText().c_str());	
      }

    | FloatConst {
        e = create_float_constant( suif_env, floatType,
                                   #expression->getText().c_str());
      }

    | DoubleConst {
        e = create_float_constant( suif_env, doubleType,
                                   #expression->getText().c_str());
      }
     
    | "true" { e = create_int_constant( suif_env, booleanType, 1); }

    | "false" { e = create_int_constant( suif_env, booleanType, 0); }

    | var:ID
      {
        VariableSymbol *vs = find_variable(c, var);
	DataType *dt = vs->get_type()->get_base_type();
        e = create_load_variable_expression( suif_env, dt, vs );
      }
    | str:StringConstant
      {
        e =
	  create_string_constant(suif_env, stringType, str->getText().c_str());
      }
    | #(DQUESTION trap:ID)
      {
        TrapSymbol *ts = find_trap(c, trap);
	DataType *dt = ts->get_type()->get_base_type();
	if (dt == voidType) {
	  report_error(trap,
		       "attempting to take value of pure exception `%s'",
		       trap->getText().c_str());
	}
	e = create_load_trap_value_expression(suif_env, dt, ts);
      }
    | #(QUESTION sig:ID)
      {
	SignalSymbol *ss = find_signal(c, sig);
	DataType *dt = ss->get_type()->get_base_type();
	if (dt == voidType) {
	  report_error(sig,
		       "attempting to take value of pure signal `%s'",
		       sig->getText().c_str());
	}
	e = create_load_signal_value_expression(suif_env, dt, ss);
      }
    | #(CALL func:ID
        {
	  LString name = func->getText().c_str();
	  SymbolTableObject *sto = c->functions->lookup_lookup_table(name);
	  if (!sto) {
	    report_error(func, "undeclared function `%s'", name.c_str());

	    // Create an integer placeholder and skip the arguments
	    e = create_int_constant(suif_env, integerType, 0);
	    break;
	  }
	  EsterelFunctionSymbol *efs = to<EsterelFunctionSymbol>(sto);
	  EsterelFunctionType *eft = efs->get_type();
	  DataType *dt = eft->get_result()->get_base_type();
	  FunctionCall *fc = create_function_call(suif_env, dt, efs);
	  Expression *arg;
        }
        ( arg=expression[c] { fc->append_argument(arg); } )*
        {
	  if ( fc->get_argument_count() != eft->get_argument_count() ) {
	    report_error(func,
	      "wrong number of arguments to call of `%s' (expected %d got %d)",
			 name.c_str(), eft->get_argument_count(),
			 fc->get_argument_count());
	  }
	  for ( size_t i = 0 ;
		i < fc->get_argument_count() && i < eft->get_argument_count() ;
		i++ ) {
	    DataType *dt1 = fc->get_argument(i)->get_result_type();
	    DataType *dt2 = eft->get_argument(i)->get_base_type();
	    if ( dt1 != dt2 ) {
	      report_error(fc->get_argument(i),
			   "type of argument %d does not match (expected `%s' got `%s')",
			   i, dt2->get_name().c_str(), dt1->get_name().c_str());
	    }
	  }
	  e = fc;
        }
      )
    // FIXME: need "pre" operator
    )
    { attach_token_linenumber(e,#expression); }
  ;

sigExpression [Context *c]
  returns [Expression *e]
  : { Expression *e1, *e2; }
    (
      #( "and" e1=sigExpression[c] e2=sigExpression[c]
        { e = create_binary_expression( suif_env, e1->get_result_type(),
					"and", e1, e2); } )
    | #( "or" e1=sigExpression[c] e2=sigExpression[c]
        { e = create_binary_expression( suif_env, e1->get_result_type(),
					"or", e1, e2); } )
    | #( "not" e1=sigExpression[c]
        { e = create_unary_expression( suif_env, e1->get_result_type(),
					"not", e1); } )
    | #( "pre" e1=sigExpression[c]
        { e = create_unary_expression( suif_env, e1->get_result_type(),
					"pre", e1); } )
    | sig:ID
      {
	SignalSymbol *ss = find_signal(c, sig);
	e = create_load_signal_expression(suif_env, booleanType, ss);
      }
    )
    { attach_token_linenumber(e,#sigExpression); }
  ;

delayExpression [Context *c]
  returns [Expression *e]
  : { Expression *e1, *e2; }
    (
      e=sigExpression[c]
    | #("immediate" e1=sigExpression[c]
      { e = create_delay(suif_env, booleanType, e1, 0, true); }
      )
    | #(DELAY e2=count:expression[c] e1=sigExpression[c]
      {
        Delay * d = create_delay(suif_env, booleanType, e1, e2);
        if ( e2->get_result_type() != integerType ) {
          report_error(count, "type of count expression must be integer, found `%s'",
		       e2->get_result_type()->get_name().c_str());
        }
        e = d;
      }
      )
    )
    { attach_token_linenumber(e,#delayExpression); }
  ;

trapExpression [Context *c] returns [Expression *e]
  : { Expression *e1, *e2;}
    ( trap:ID { TrapSymbol *ts = find_trap(c, trap);
                e = create_load_trap_expression(suif_env, booleanType, ts); }
    | #( "and" e1=trapExpression[c] e2=trapExpression[c]
        { e = create_binary_expression( suif_env, booleanType,
					"and", e1, e2 ); } )
    | #( "or" e1=trapExpression[c] e2=trapExpression[c]
        { e = create_binary_expression( suif_env, booleanType,
					"or", e1, e2 ); } )
    | #( "not" e1=trapExpression[c]
         { e = create_unary_expression( suif_env, booleanType, "not", e1 ); } )
    )
  ;

procArgs [Context *c, ProcedureCallStatement *pcs]
  : #( VARS
      ( var:ID { pcs->append_reference_arg(find_variable(c, var)); } )*
    )
    #( ARGS { Expression *e; }
       ( e=expression[c] { pcs->append_value_arg(e); } )*
    )
    // FIXME: Typecheck this bad boy
  ;

taskCall [Context *c, ExecStatement *es]
  : #( cas:CASE task:ID
       {
	 LString name = task->getText().c_str();
	 suif_assert(c->tasks);
	 SymbolTableObject *sto = c->tasks->lookup_lookup_table(name);
	 if (!sto) {
	   report_error(task, "undeclared task `%s'", name.c_str());
	   return;
	 }
	 suif_assert(sto);
	 EsterelProcedureSymbol *eps = to<EsterelProcedureSymbol>(sto);
	 suif_assert(eps);
	 EsterelProcedureType *ept = eps->get_type();
	 TaskCall *tc = create_task_call(suif_env, eps, 0);
	 es->append_call(tc);
	 attach_token_linenumber(tc, cas);
	 suif_assert(ept);
       }
       procArgs[c, tc]
       sig:ID
       {
	 SignalSymbol *ss = find_signal(c, sig);
	 tc->set_signal(ss);
	 Type *t = ss->get_type();
	 suif_assert(t);
	 SignalType *st = to<SignalType>(t);
	 if ( st->get_direction() != returnAnnote ) {
	   report_error(sig, "%s signal `%s' not of type `return'",
			st->get_direction().c_str(),
			ss->get_name().c_str());
	   report_warning(ss, "`%s' declared here", ss->get_name().c_str() );
	 }
	 Statement *s;
       }
       ( s=statement[c] { tc->set_body(s); } )?
     )
  ;

typeRenaming [Context *c, RunStatement *rs]
 : #( TRENAME { QualifiedType *qt; } qt=newType:type[c] oldID:ID
     {
       TypeRenaming *tr =
	 create_type_renaming(suif_env, oldID->getText().c_str(), qt);
       attach_token_linenumber(tr, newType);
       rs->append_type_renaming(tr);
     }
   )
 ;    

constantRenaming [Context *c, RunStatement *rs]
 : #( CRENAME { Expression *e; } e=expression[c] oldID:ID
     {
       ConstantRenaming *cr =
	 create_constant_renaming(suif_env, oldID->getText().c_str(), e);
       attach_token_linenumber(cr, #constantRenaming);
       rs->append_constant_renaming(cr);
     }
   )
 ;    

functionRenaming [Context *c, RunStatement *rs]
 : #( FRENAME { EsterelFunctionSymbol *efs; }
     efs=func:newFunctionName[c]
     oldID:ID
     {
       if (efs) {
	 FunctionRenaming *fr =
	   create_function_renaming(suif_env, oldID->getText().c_str(), efs);
	 attach_token_linenumber(fr, func);
	 suif_assert(rs);
	 rs->append_function_renaming(fr);
       }
     }
   )
 ;

newFunctionName [Context *c] returns [EsterelFunctionSymbol *efs]
  : { efs = 0; }
    ( func:ID
       {
	 LString name = func->getText().c_str();
	 SymbolTableObject *sto = c->functions->lookup_lookup_table(name);
	 if (!sto) {
	   report_error(func, "undeclared function `%s'", name.c_str());
	 } else {
	   efs = to<EsterelFunctionSymbol>(sto);
	 }
       }
    | pf:predefinedFunction
      { efs = get_builtin_function(c, pf->getText().c_str()); }
    )
  ;

predefinedFunction
  : "and"
  | "or"
  | "not"
  | PLUS
  | DASH
  | STAR
  | SLASH
  | "mod"
  | LESSTHAN
  | GREATERTHAN
  | LEQUAL
  | GEQUAL
  | NEQUAL
  | EQUALS
  ;

procedureRenaming [Context *c, RunStatement *rs]
 : #( PRENAME proc:ID oldID:ID
     {
       LString name = proc->getText().c_str();
       SymbolTableObject *sto = c->procedures->lookup_lookup_table(name);
       if (!sto) {
	 report_error(proc, "undeclared procedure `%s'", name.c_str());
       } else {
	 EsterelProcedureSymbol *eps = to<EsterelProcedureSymbol>(sto);
	 ProcedureRenaming *pr =
	   create_procedure_renaming(suif_env, oldID->getText().c_str(), eps);
	 attach_token_linenumber(pr, proc);
	 rs->append_procedure_renaming(pr);
       }
     }
   )
 ;

taskRenaming [Context *c, RunStatement *rs]
 : #( TARENAME proc:ID oldID:ID
     {
       LString name = proc->getText().c_str();
       SymbolTableObject *sto = c->tasks->lookup_lookup_table(name);
       if (!sto) {
	 report_error(proc, "undeclared task `%s'", name.c_str());
       } else {
	 EsterelProcedureSymbol *eps = to<EsterelProcedureSymbol>(sto);
	 ProcedureRenaming *tr =
	   create_procedure_renaming(suif_env, oldID->getText().c_str(), eps);
	 attach_token_linenumber(tr, proc);
	 rs->append_task_renaming(tr);
       }
     }
   )
 ;

signalRenaming [Context *c, RunStatement *rs]
 : #( SRENAME sig:ID oldID:ID
     {
       SignalSymbol *ss = find_signal(c, sig);
       SignalRenaming *sr =
	   create_signal_renaming(suif_env, oldID->getText().c_str(), ss);
       attach_token_linenumber(sr, sig);
       rs->append_signal_renaming(sr);
     }
   )
 ;
