#include <basicnodes/basic.h>
#include <suifnodes/suif.h>
#include <esuifnodes/esuif.h>
#include <common/MString.h>
#include <common/lstring.h>
#include <common/suif_map.h>
#include <suifkernel/visitor_map.h>
#include <suifkernel/suif_env.h>

#define VISITORS \
  VISITOR(NothingStatement); \
  VISITOR(HaltStatement); \
  VISITOR(PauseStatement); \
  VISITOR(StatementList); \
  VISITOR(ParallelStatementList); \
  VISITOR(StoreVariableStatement); \
  VISITOR(LoadVariableExpression); \
  VISITOR(LoadTrapExpression); \
  VISITOR(LoadTrapValueExpression); \
  VISITOR(LoadSignalExpression); \
  VISITOR(LoadSignalValueExpression); \
  VISITOR(Delay); \
  VISITOR(IntConstant); \
  VISITOR(StringConstant); \
  VISITOR(FloatConstant); \
  VISITOR(BinaryExpression); \
  VISITOR(UnaryExpression); \
  VISITOR(VariableStatement); \
  VISITOR(TrapStatement); \
  VISITOR(SignalStatement); \
  VISITOR(EmitStatement); \
  VISITOR(ExitStatement); \
  VISITOR(SustainStatement); \
  VISITOR(PresentStatement); \
  VISITOR(EsterelIfStatement); \
  VISITOR(LoopStatement); \
  VISITOR(LoopEachStatement); \
  VISITOR(EveryStatement); \
  VISITOR(SuspendStatement); \
  VISITOR(DoWatchingStatement); \
  VISITOR(DoUptoStatement); \
  VISITOR(RepeatStatement); \
  VISITOR(AbortStatement); \
  VISITOR(ProcedureCallStatement); \
  VISITOR(ExecStatement); \
  VISITOR(FunctionCall); \
  VISITOR(AwaitStatement); \
  VISITOR(RunStatement)


class EsterelPrinter {
public:

  static const int sequentialPrecedence = 2;
  static const int parallelPrecedence = 1;

  EsterelPrinter(SuifEnv*);
  ~EsterelPrinter();
  VisitorMap *map;

  class State {
  public:
    State() : _indent(0), _precedence(0), _nextState(0) {}

    State( State *st ) {
      suif_assert(st);
      _indent = st->_indent;
      _precedence = st->_precedence;
      _nextState = st;
    }

    State * get_next_state() const { return _nextState; }
    int get_indent() const { return _indent; }
    void indent() { _indent += 1; }
    int get_precedence() const { return _precedence; }
    bool set_precedence(int p) {
      bool needBrackets = p < _precedence;
      _precedence = p;
      return needBrackets;
    }

    void add_indent(String &result);

  private:
  
    int _indent;	// Current indentation -- used by add_indent()
    int _precedence;	// Precedence for statements or expressions
    State *_nextState;
  };

  State *state;

  void pop_state() { state = state->get_next_state(); }
  void indent() { state->add_indent(result); }

  void expression(Expression *);
  void sigExpression(Expression *);
  void statement(Statement *);

  /// All functions append text to the result String
  String result;

  void fileSetBlock(const FileSetBlock *);
  void module(const ModuleDefinition *);
  void types(const ModuleDefinition *);
  void constants(const ModuleDefinition *);
  void functions(const ModuleDefinition *);
  void procedures(const ModuleDefinition *);
  void tasks(const ModuleDefinition *);
  void signals(const ModuleDefinition *);
  void relations(const ModuleDefinition *);

  void taskCall(const TaskCall *);

#define VISITOR(class) \
  static void visit ## class(void *info, Object *obj) { \
    ((EsterelPrinter *)info)->print ## class((class *)obj); \
  } \
  void print ## class(const class *);

  VISITORS;
#undef VISITOR

  static void visitUnknown(void *info, Object *obj) {
    ((EsterelPrinter *)info)->printUnknown(obj);
  }
  void printUnknown(Object *);

  int op_precedence(LString);

private:
  suif_map<LString, int> _operatorPrecedence;

};
